/*******************************************************************************
 * Copyright (c) 2013-2015 Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser Public License v3
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl.html
 *
 * Contributors:
 *     Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza - initial API and implementation
 ******************************************************************************/
package com.net2plan.examples.netDesignAlgorithm.cfa;

import cern.colt.matrix.tdouble.DoubleFactory2D;
import cern.colt.matrix.tdouble.DoubleMatrix2D;
import com.jom.OptimizationProblem;
import com.net2plan.interfaces.networkDesign.IAlgorithm;
import com.net2plan.interfaces.networkDesign.Net2PlanException;
import com.net2plan.interfaces.networkDesign.NetPlan;
import com.net2plan.libraries.GraphUtils;
import com.net2plan.utils.CollectionUtils;
import com.net2plan.utils.Constants;
import com.net2plan.utils.Pair;
import com.net2plan.utils.Triple;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Given a network topology, and the offered traffic (measured in Gbps), this 
 * algorithm obtains the traffic routing and the (modular) capacities in the links 
 * that minimizes the link costs. The capacity of a link is constrained to be the 
 * aggregation of integer multiples of modules of capacities {0.15, 0.6, 2.4, 9.6} Gbps, 
 * and prices {1, 2, 4, 8} monetary units. Link utilization is limited by the 
 * user-defined parameter rhoMax.
 *
 * @author Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza
 * @version 1.2, September 2015
 */
public class CFA_minCostModularCapacities_xde implements IAlgorithm
{
	@Override
	public String executeAlgorithm(NetPlan netPlan, Map<String, String> algorithmParameters, Map<String, String> net2planParameters)
	{
		/* Basic checks */
		int N = netPlan.getNumberOfNodes();
		int E = netPlan.getNumberOfLinks();
		int D = netPlan.getNumberOfDemands();
		if (N == 0 || E == 0 || D == 0) throw new Net2PlanException("This algorithm requires a topology with links, and a demand set");
		
		/* Initialize some variables */
		double[] h_d_Gbps = netPlan.getDemandOfferedTrafficVector();
		double[] p_i = new double[] { 1, 2, 4, 8 };
		double[] u_i_Gbps = new double[] { 0.15, 0.6, 2.4, 9.6 };

		/* Take some user-defined parameters */
		double rhoMax = Double.parseDouble(algorithmParameters.get("rhoMax"));
		if ((rhoMax <= 0) || (rhoMax > 1)) throw new Net2PlanException("rhoMax parameter must be positive, and equal or lower than one");

		/* Create the optimization problem object (JOM library) */
		OptimizationProblem op = new OptimizationProblem();

		/* Add the decision variables to the problem */
		op.addDecisionVariable("x_de", false, new int[] { D, E }, 0, 1);
		op.addDecisionVariable("y_e", false, new int[] { 1, E }, 0, Double.MAX_VALUE);
		op.addDecisionVariable("n_ei", true, new int[] { E, p_i.length }, 0, Double.MAX_VALUE);

		/* Set some input parameters */
		op.setInputParameter("p_i", p_i, "row");
		op.setInputParameter("u_i", u_i_Gbps, "row");
		op.setInputParameter("h_d_Gbps", h_d_Gbps, "row");
		op.setInputParameter("rhoMax", rhoMax);

		/* Sets the objective function */
		op.setObjectiveFunction("minimize", "sum(n_ei * p_i')");

		/* VECTORIAL FORM OF THE CONSTRAINTS */
		Set<Long> nodeIds = netPlan.getNodeIds();
		Map<Long, Pair<Long, Long>> linkMap = netPlan.getLinkMap();
		Map<Long, Pair<Long, Long>> demandMap = netPlan.getDemandMap();
		op.setInputParameter("A_e", GraphUtils.getIncidenceMatrix(nodeIds, linkMap));
		op.setInputParameter("A_d", GraphUtils.getIncidenceMatrix(nodeIds, demandMap));
		op.addConstraint("A_e * x_de' == A_d"); /* the flow-conservation constraints (NxD constraints) */
		op.addConstraint("y_e == h_d_Gbps * x_de"); /* computing the traffic in each link (E constraints) */
		op.addConstraint("y_e <= rhoMax * u_i * n_ei'"); /* the utilization in the link is below rhoMax */

		/* Call the solver to solve the problem */
		String solverName = algorithmParameters.get("solverName");
		String solverLibraryName = algorithmParameters.get("solverLibraryName");
		op.solve(solverName, "solverLibraryName", solverLibraryName);

		/* If an optimal solution was not found, quit */
		if (!op.solutionIsOptimal()) throw new Net2PlanException("An optimal solution was not found");

		/* Retrieve the optimum solutions */
		DoubleMatrix2D n_ei = op.getPrimalSolution("n_ei").view2D();
		DoubleMatrix2D x_de = op.getPrimalSolution("x_de").view2D();
		x_de = DoubleFactory2D.sparse.diagonal(h_d_Gbps).zMult(x_de, null);

		/* Update netPlan object: routing */
		netPlan.removeAllRoutingInformation();
		netPlan.setRoutingType(Constants.RoutingType.SOURCE_ROUTING);
		netPlan.addRoutesFromDemandLinkCarriedTraffic(x_de);

		/* Update netPlan object: link capacities */
		Set<Long> linkIds = netPlan.getLinkIds();
		Map<Long, Integer> linkId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(linkIds);
		for (long linkId : linkIds)
		{
			int e = linkId2LinearIndex.get(linkId);
			double thisLinkCapacity_Gbps = 0;
			for (int capacityModuleId = 0; capacityModuleId < u_i_Gbps.length; capacityModuleId++)
				thisLinkCapacity_Gbps += u_i_Gbps[capacityModuleId] * n_ei.getQuick(e, capacityModuleId);
			
			netPlan.setLinkCapacity(linkId, thisLinkCapacity_Gbps);
		}

		/* Check that the budget is used */
		return "Ok! Cost: " + op.getOptimalCost();
	}

	@Override
	public String getDescription()
	{
		return "Given a network topology, and the offered traffic (measured in Gbps), "
				+ "this algorithm obtains the traffic routing and the (modular) "
				+ "capacities in the links that minimizes the link costs. The "
				+ "capacity of a link is constrained to be the aggregation of "
				+ "integer multiples of modules of capacities {0.15, 0.6, 2.4,"
				+ " 9.6} Gbps, and prices {1, 2, 4, 8} monetary units. Link "
				+ "utilization is limited by the user-defined parameter rhoMax";
	}

	@Override
	public List<Triple<String, String, String>> getParameters()
	{
		List<Triple<String, String, String>> algorithmParameters = new LinkedList<Triple<String, String, String>>();
		algorithmParameters.add(Triple.of("rhoMax", "0.5", "Maximum utilization allowed in the links"));
		algorithmParameters.add(Triple.of("solverName", "glpk", "The solver name to be used by JOM"));
		algorithmParameters.add(Triple.of("solverLibraryName", "", "The solver library full or relative path, to be used by JOM. Leave blank to use JOM default."));
		
		return algorithmParameters;
	}
}